import torch
from typing import TYPE_CHECKING
from pathlib import Path
from datatia.utils import longpath
if TYPE_CHECKING:
    from datatia.actions import Truncate, PreMap

class TensorSource:
    def __init__(self, path : str,
            this_truncate_actions : list['Truncate'] = [],
            this_premap_actions : list['PreMap'] = [],
            keep_in_memory : bool = False):
        self.path = Path(path)
        self.keep_in_memory = keep_in_memory

        self.this_truncate_actions = this_truncate_actions
        if not self.path.exists():
            raise ValueError(f"TensorProxy path {self.path} does not exist")

        if self.keep_in_memory:
            self.tensor = self.truncate(torch.load(longpath(self.path), map_location='cpu'))
            self.tensor = self.premap(this_premap_actions, self.tensor)

    def premap(self,
            this_premap_actions : list['PreMap'],
            tensor : torch.Tensor):
        for action in (this_premap_actions):
            tensor = action.apply(tensor)
        return tensor

    def truncate(self, tensor):
        for action in (self.this_truncate_actions):
            tensor = action.apply(tensor)
        return tensor

    def get(self):
        if self.keep_in_memory:
            return self.tensor
        else:
            return self.truncate(torch.load(longpath(self.path)))


import os
def longpath(path):
    import platform
    path = os.path.abspath(path)
    if 'Windows' in platform.system() and not path.startswith('\\\\?\\'):
        path = u'\\\\?\\'+path.replace('/','\\')
        return path
    else:
        return path


import datatia as dt
import torch
import os

TEST_FILES_DIR = 'test/test_files'
TEST_FILES_LIST = 'test/filelist.txt'

def generate_dummy_files():
    """
    Generate a set of dummy files for testing datatia.

    This function writes a bunch of random tensors to disk, and creates a filelist
    that points to them. The filelist is written to the file specified by
    TEST_FILES_LIST.

    The dummy files are written to the directory specified by TEST_FILES_DIR.

    The filelist has the following format:

    tensor1_0.pt|tensor2_0.pt|0
    tensor1_1.pt|tensor2_1.pt|1
    tensor1_2.pt|tensor2_2.pt|2
    tensor1_3.pt|tensor2_3.pt|3

    The first column is the path to the first tensor, the second column is the
    path to the second tensor, and the third column is the label.

    The tensors are all of shape (8, 32) for tensor1 and (10, 32) for tensor2, 
    and the labels are all integers from 0 to 7, 
    EXCEPT the first tensor1 which is of shape (4, 32).

    """
    os.makedirs(TEST_FILES_DIR, exist_ok=True)
    
    filelist = []
    for i in range(8):
        if i == 0:
            tensor1 = torch.randn([4, 32])
        else:
            tensor1 = torch.randn([8, 32])
        tensor2 = torch.randn([10, 32])

        tensor1_path = os.path.join(TEST_FILES_DIR, f'tensor1_{i}.pt')
        torch.save(tensor1, tensor1_path)
        tensor2_path = os.path.join(TEST_FILES_DIR, f'tensor2_{i}.pt')
        torch.save(tensor2, tensor2_path)

        line = f'{tensor1_path}|{tensor2_path}|{i}'
        filelist.append(line)

    with open(TEST_FILES_LIST, 'w') as f:
        f.write('\n'.join(filelist))

def test_datatia():
    generate_dummy_files()
    field_specs = [dt.FieldSpec(
            name='tensor1', datatype=torch.Tensor, provide_length=True),
            dt.FieldSpec(
            name='tensor2', datatype=torch.Tensor, keep_in_memory=False),
            dt.FieldSpec(
            name='label', datatype=int)]

    # Test dataset loading and truncate
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.Truncate(field='tensor1', dims=[0], max_lengths=[4])])
    assert dataset[0]['tensor1'].shape == torch.Size([4, 32])
    assert dataset[0]['tensor2'].shape == torch.Size([10, 32])

    # Test dropping
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.Drop(condition=lambda row: row['label'] == 0)])
    assert len(dataset) == 7

    # Test premap on literal field
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.PreMap(field='label', operation=lambda x: x + 1)])
    assert dataset[0]['label'] == 1

    # Test premap on tensor field
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.PreMap(field='tensor1', operation=lambda x: x * 0)])
    assert dataset[0]['tensor1'].abs().sum() == 0

    # Test live map row
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.LiveMapRow(operation=lambda row: {k: v * 0 for k, v in row.items()})])
    assert dataset[1]['tensor1'].abs().sum() == 0
    assert dataset[1]['tensor2'].abs().sum() == 0
    assert dataset[1]['label'] == 0

    # Test random subsample
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.RandomSubsample(fields=['tensor1', 'tensor2'], 
            dims=[0], length=4)])
    assert dataset[0]['tensor1'].shape[0] == 4
    assert dataset[0]['tensor2'].shape[0] == 4

    # Test pad group
    dataset = dt.Dataset(filelist=TEST_FILES_LIST,
        field_specs = field_specs,
        actions=[dt.PadGroup(fields=['tensor1', 'tensor2'], 
            dims=[0, 1], values=[0, 0], to_multiple=[4, 5])])
    loader = dataset.loader(batch_size=4)
    batch = next(iter(loader))
    assert batch['tensor1'][0].shape[0] == 12
    assert batch['tensor1'][0].shape[1] == 35
    assert batch['tensor2'][0].shape[0] == 12
    assert batch['tensor2'][0].shape[1] == 35
    assert batch['tensor1'][0][8:, :].abs().sum() == 0
    assert batch.get('tensor1_length') is not None
    assert batch.get('tensor1_length')[0] == 4
    assert batch['label'][0] == 0
    assert batch['label'].shape == torch.Size([4])



from torch.utils.data import Dataset as torchDataset, DataLoader
from typing import Union
from pathlib import Path
from dataclasses import dataclass
from datatia.actions import (
    Action, Truncate, Drop, PreMap, FieldData, LiveMapRow, RandomSubsample, PadGroup,
    RowData)
from datatia.tensor_source import TensorSource
import torch

@dataclass
class FieldSpec:
    name : str
    datatype : type
    dim : Union[torch.Size, None] = None # -1 should specify a variable dimension

    # For use with PadGroup only: create an extra output field with name {name}_length
    provide_length : bool = False 

    keep_in_memory : bool = True

    def validate(self):
        """
        Validate a FieldSpec

        This checks that all the fields are valid: name is not empty, datatype is one of the allowed types,
        and dim is either None or a torch.Size.

        Raises:
            ValueError: if the FieldSpec is invalid
        """
        if self.name is None or len(self.name) == 0:
            raise ValueError(f"FieldSpec {self.name} has no name")
        allowed_types = [str, int, float, bool, torch.Tensor]
        if self.datatype not in allowed_types:
            raise ValueError(f"FieldSpec {self.name} has invalid datatype {self.datatype}")
        if self.dim is not None and not isinstance(self.dim, torch.Size):
            raise ValueError(f"FieldSpec {self.name} has invalid dim {self.dim}")
        if not self.keep_in_memory and not self.datatype is torch.Tensor:
            raise ValueError(f"keep_in_memory=False is only supported for torch.Tensor datatypes")
        if not self.datatype is torch.Tensor and self.dim is not None:
            raise ValueError(f"FieldSpec {self.name} has dim {self.dim} but datatype is not torch.Tensor")

    def data_validate(self, data):
        """
        Validate the data against the FieldSpec.

        This function checks if the datatype of the provided data matches the
        specified datatype of the FieldSpec. If the datatype is torch.Tensor,
        it also verifies that the shape of the data matches the specified dim
        field, considering -1 as a wildcard dimension.

        Args:
            data: The data to validate against the FieldSpec.

        Raises:
            ValueError: If the datatype of data does not match the FieldSpec's
            datatype, or if data is a torch.Tensor and its shape does not match
            the specified dim.
        """

        if type(data) is not self.datatype:
            raise ValueError(f"FieldSpec {self.name} has datatype {self.datatype} but data is of type {type(data)}")
        if type(data) is torch.Tensor and self.dim is not None:
            if not self.shape_matches(data):
                raise ValueError(f"FieldSpec {self.name} has dim {self.dim} but data has shape {data.shape}")

    def shape_matches(self, data):
        """
        Check that the shape of a tensor matches the specified dim field of the FieldSpec.

        If dim is None, this function always returns True. If dim is a torch.Size, this
        function checks that the shape of the tensor matches this, ignoring any dimensions
        that are -1 in the FieldSpec.

        Args:
            data: the tensor to check

        Returns:
            True if the shape matches, False otherwise
        """
        for dim in self.dim:
            if dim == -1:
                continue
            if data.shape[dim] != self.dim[dim]:
                return False
        return True

class Dataset(torchDataset):
    def __init__(self,
        filelist: Union[str, Path, list[str]],
        field_specs: list[FieldSpec],
        is_train: bool = True, # controls training-only actions like random subsampling
        actions: list[Action] = [],
        *args, **kwargs):
        """
        Create a new Dataset.

        Args:
            filelist: path to filelist file, or a list of filelist entries
            field_specs: list of FieldSpecs describing each column in the filelist
            is_train: whether to apply training-only actions like random subsampling
            actions: list of Actions to apply to data
            args/kwargs: passed to torch.utils.data.Dataset.__init__

        Raises:
            ValueError: if the number of fields in the filelist does not match the number of FieldSpecs

        """
        super().__init__(*args, **kwargs)

        self.validate_specs(field_specs)
        filelist = self.process_filelist_file(filelist)
        filelist = self.create_filelist_rows(filelist)

        if len(filelist) and len(filelist[0]) != len(field_specs):
            raise ValueError(f"Number of fields in filelist ({len(filelist[0])}) does not match number of FieldSpecs ({len(field_specs)})")

        self.actions = actions

        self.live_actions = {
            'live_map_row': [ action for action in actions if isinstance(action, LiveMapRow) ],
            'random_subsample': [ action for action in actions if isinstance(action, RandomSubsample) ],
            'pad_group': [ action for action in actions if isinstance(action, PadGroup) ]
        }

        self.load_data(filelist, field_specs)
        self.field_specs = field_specs
        self.field_specs_map = { field_spec.name: field_spec for field_spec in field_specs }
        self.is_train = is_train

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data = self.data[idx]
        # Resolve any TensorSources
        data : RowData = {
            k: v.get() if isinstance(v, TensorSource) else v for k, v in data.items()
        }
        for action in self.live_actions['live_map_row']:
            data = action.apply(data)
        if self.is_train:
            for action in self.live_actions['random_subsample']:
                field_names = action.fields
                field_datas = [ data[field_name] for field_name in field_names ]
                new_tensors = action.apply(field_datas)
                for i, field_name in enumerate(field_names):
                    data[field_name] = new_tensors[i]
        return data

    def collate_fn(self, batch):
        ret = {}
        for action in self.live_actions['pad_group']:
            field_names = action.fields
            batch_tensors = [ [ data[field_name] for field_name in field_names ] for data in batch ]
            new_batch_tensors, unpadded_lengths = action.apply(batch_tensors)
            for row in range(len(batch)):
                for i, field_name in enumerate(field_names):
                    batch[row][field_name] = new_batch_tensors[row][i]
                    if self.field_specs_map[field_name].provide_length:
                        batch[row][f'{field_name}_length'] = unpadded_lengths[field_name][row]

        for field_spec in self.field_specs:
            if field_spec.datatype is torch.Tensor:
                ret[field_spec.name] = torch.stack([d[field_spec.name] for d in batch])
                if field_spec.provide_length:
                    ret[f'{field_spec.name}_length'] = torch.tensor([d[f'{field_spec.name}_length'] for d in batch])
            elif field_spec.datatype is int:
                ret[field_spec.name] = torch.tensor([d[field_spec.name] for d in batch])
            else:
                ret[field_spec.name] = [d[field_spec.name] for d in batch]

        return ret

    def loader(self, *args, **kwargs):
        return DataLoader(self, collate_fn=self.collate_fn, *args, **kwargs)

    def load_data(self, 
        filelist: Union[str, Path, list[str]],
        field_specs: list[FieldSpec]):
        """
        Load the data from the filelist into memory
        and apply Truncate and Drop actions.

        For each row in the filelist, and for each field in the row, this
        method creates a dictionary mapping the field name to the value.
        The value is converted to the type specified in the FieldSpec.

        If the FieldSpec.datatype is torch.Tensor, then the value is treated as
        a path to a torch tensor, and the tensor is loaded from disk. If
        FieldSpec.keep_in_memory is true, the tensor is loaded immediately and
        stored in memory. If FieldSpec.keep_in_memory is false, the path to the
        tensor is stored instead.

        The resulting list of dictionaries is stored in the self.data attribute.

        Args:
            filelist: the filelist to load data from
            field_specs: the list of FieldSpecs to use to interpret the filelist
        """
        self.data = []

        self.field_specs_map = {
            field_spec.name: field_spec for field_spec in field_specs
        }

        self.drop_actions = [
            action for action in self.actions if isinstance(action, Drop)
        ]
        self.truncate_actions = {
            field: [] for field in [field_spec.name for field_spec in field_specs]
        }
        self.premap_actions = {
            field: [] for field in [field_spec.name for field_spec in field_specs]
        }

        for action in self.actions:
            if isinstance(action, Truncate):
                action : Truncate
                if self.field_specs_map[action.field].datatype != torch.Tensor:
                    raise ValueError(f"Truncate action on non-tensor field {action.field}")
                self.truncate_actions[action.field].append(action)
            if isinstance(action, PreMap):
                if self.field_specs_map[action.field].keep_in_memory == False:
                    raise ValueError(f"PreMap action on non-memory field {action.field}")
                self.premap_actions[action.field].append(action)

        for i, row in enumerate(filelist):
            data : RowData = {}
            if not len(row) == len(field_specs):
                raise ValueError(f"Number of fields in row {i} ({len(row)}) does not match number of FieldSpecs ({len(field_specs)}): {row}")
            for field_spec, value in zip(field_specs, row):
                if field_spec.datatype == torch.Tensor:
                    # Treat the text value as a path to a torch tensor
                    value = TensorSource(value, 
                        self.truncate_actions[field_spec.name],
                        self.premap_actions[field_spec.name],
                        field_spec.keep_in_memory)
                    data[field_spec.name] = value
                else:
                    data[field_spec.name] = field_spec.datatype(value)
                    data[field_spec.name] = self.premap(field_spec.name, data[field_spec.name])

            do_append = True
            for action in self.drop_actions:
                if action.test(data):
                    do_append = False
                    break

            if do_append:
                self.data.append(data)

    def premap(self, field_name : str, value : FieldData):
        for action in self.premap_actions[field_name]:
            value = action.apply(value)
        return value

    def validate_specs(self,
        field_specs: list[FieldSpec]):
        """
        Validate all FieldSpecs in a list

        This calls the validate() method on each FieldSpec in the list.

        Args:
            field_specs: list of FieldSpecs to validate

        Raises:
            ValueError: if any FieldSpec is invalid
        """
        for field_spec in field_specs:
            field_spec.validate()

    def create_filelist_rows(self,
        true_filelist: list[str],
        delimiter: str = '|'):
        """
        Split a list of strings into a list of lists of strings, split by delimiter.

        Args:
            true_filelist: the list of strings to split
            delimiter: the delimiter to split the strings with

        Returns:
            A list of lists of strings, where each inner list is the split version of the corresponding input string.
        """
        rows = []
        for line in true_filelist:
            line = line.removesuffix('\n')
            rows.append([e for e in line.split(delimiter)])
        return rows

    def process_filelist_file(self,
        filelist: Union[str, Path, list[str]]):
        """
        Convert a filelist input into a list of strings.

        Args:
            filelist: the filelist, which can be either a string (interpreted as a
                path to a filelist file), a Path object (interpreted as a path to a
                filelist file), or a list of strings (interpreted as the filelist)

        Returns:
            a list of strings, which is the filelist
        """
        
        if type(filelist) is list:
            return filelist
        if type(filelist) is str:
            filelist_path = Path(filelist)
            if not filelist_path.exists():
                raise ValueError(f"Filelist path {filelist_path} does not exist")
        filelist : Path
        with open(filelist, encoding='utf-8') as f:
            lines = f.readlines()
        return lines



import math
import torch
from typing import Union, Callable, TYPE_CHECKING
from datatia.tensor_source import TensorSource
from enum import Enum

FieldData = Union[str, int, float, bool, torch.Tensor]
FieldDataProxied = Union[str, int, float, bool, TensorSource]
RowData = dict[str, FieldData]
RowDataProxied = dict[str, FieldDataProxied]

class Action:
    """
    Base class for all actions
    """
    def __init__(self):
        pass

    def resolveRow(self, row : RowDataProxied) -> RowData:
        return {
            k: v.get() if isinstance(v, TensorSource) else v for k, v in row.items()}

class Truncate(Action):
    """
    Truncate a tensor field from the beginning. 
    Applied on first data load (or live if keep_in_memory is False).
    Truncation is always performed before any other action.
    """
    def __init__(self, field : str, 
        dims : list[int], max_lengths : list[int]):
        super().__init__()
        self.field = field
        self.dims = dims
        self.max_lengths = max_lengths

    def apply(self, tensor : torch.Tensor):
        for dim, max_length in zip(self.dims, self.max_lengths):
            tensor = tensor.narrow(dim, 0, max_length)
        return tensor

class Drop(Action):
    """
    Drop a row if a condition is met (True means drop the row).
    Applied only on first data load
    """
    def __init__(self, condition : Callable[[RowData], bool]):
        super().__init__()
        self.condition = condition

    def test(self, row : RowDataProxied):
        return self.condition(self.resolveRow(row))

class PreMap(Action):
    """
    Apply an operation on a field value on first data load
    (cannot be used if keep_in_memory is False)
    """
    def __init__(self, 
        field : str, 
        operation : Callable[[FieldData], FieldData]):
        super().__init__()
        self.field = field
        self.operation = operation

    def apply(self, field : FieldData):
        return self.operation(field)

class LiveMapRow(Action):
    """
    Apply an operation on a row on every data access 
    (i.e. for random data augmentation)
    """
    def __init__(self, 
        operation : Callable[[RowData], RowData]):
        super().__init__()
        self.operation = operation

    def apply(self, row : RowData):
        return self.operation(row)

class RandomSubsample(Action):
    """
    Randomly subsample sequence dimensions of one or more tensor fields
    on every data access, using the same starting index across all fields.
    Use frame_multiples for frame-aligned features (e.g., spectrogram vs waveform)
    Only runs if is_train=True for the dataset.
    """
    def __init__(self,
        fields: list[str],
        dims: list[int],
        length: int,  # Single length since all fields represent same time duration
        frame_multiples: list[int] = []):
        super().__init__()
        self.fields = fields 
        self.dims = dims
        self.length = length
        if len(frame_multiples) == 0:
            self.frame_multiples = [1 for _ in range(len(fields))]
        else:
            self.frame_multiples = frame_multiples
    
    def apply(self, field_tensors: list[torch.Tensor]):
        if not field_tensors:
            return field_tensors
        
        new_tensors = []
        
        for i, tensor in enumerate(field_tensors):
            dim = self.dims[i] if i < len(self.dims) else self.dims[-1]
            frame_mult = self.frame_multiples[i]
            
            target_length = self.length * frame_mult
            target_length = min(target_length, tensor.shape[dim])
            tensor_length = tensor.shape[dim]
            
            if tensor_length <= target_length:
                start_idx = 0
            else:
                # Ensure start_idx is aligned to frame boundaries
                max_start = (tensor_length - target_length) // frame_mult
                start_idx = torch.randint(0, max_start + 1, (1,))[0] * frame_mult
            
            new_tensor = tensor.narrow(dim, start_idx, target_length)
            new_tensors.append(new_tensor)
        
        return new_tensors

class PadGroupMode(Enum):
    UNDEFINED = 0
    MAXIMUM = 1
    MULTIPLE = 2
    LENGTH = 3

class PadGroup(Action):
    """
    For batched data, right-pad dimensions of one or more tensor fields to the
    same length per dimension using fill values.

    This assumes that the padded dimensions are the same across all fields.

    If LENGTH is specified, this assumes that the dimension to be padded is smaller or equal to the specified length.
    Use Truncate to ensure that the dimension to be padded is smaller than the specified length.

    Dimension indexing uses unbatched dimensions.
    For example, if we expect a 3D tensor with shape [batch, height, width],
    and we want to pad the height and width dimensions to the same length,
    we should specify dims=[0, 1] and to_length=[height, width].

    If neither to_multiple or to_length are specified, use the maximum length.
    """
    def __init__(self,
        fields : list[str],
        dims : list[int],
        values : list[Union[float, int]],
        to_multiple : list[Union[int, None]] = [],
        to_length : list[Union[int, None]] = [],
        ):
        super().__init__()
        self.fields = fields
        self.dims = dims
        self.to_multiple = to_multiple
        self.to_length = to_length
        self.values = values

    def determine_pad_mode(self, i):
        if len(self.to_multiple) > i and self.to_multiple[i] is not None:
            if len (self.to_length) > i and self.to_length[i] is not None:
                raise ValueError(
                    f"PadGroup cannot have both to_multiple and to_length specified at dim {i}")
            return PadGroupMode.MULTIPLE
        elif len(self.to_length) > i and self.to_length[i] is not None:
            return PadGroupMode.LENGTH
        else:
            return PadGroupMode.MAXIMUM

    def apply(self, batch_tensors : list[list[torch.Tensor]]):
        # [ [field1, field2] , [field1, field2] ]
        if not len(batch_tensors) or not len(batch_tensors[0]):
            return batch_tensors

        # 1. Determine target padding length
        target_lengths = [0 for _ in range(len(self.dims))]
        unpadded_lengths = {}

        for i in range(len(self.dims)):
            pad_group_mode = self.determine_pad_mode(i)
            if pad_group_mode == PadGroupMode.LENGTH:
                target_lengths[i] = self.to_length[i]
            else:
                for batch in batch_tensors:
                    for field_tensor in batch:
                        target_lengths[i] = max(
                            field_tensor.shape[self.dims[i]], target_lengths[i])
                if pad_group_mode == PadGroupMode.MULTIPLE:
                    target_lengths[i] = int(
                        math.ceil(target_lengths[i] / self.to_multiple[i]) * self.to_multiple[i])

        for i,field in enumerate(self.fields):
            unpadded_lengths[field] = [
                batch_tensors[j][i].shape[self.dims[i]] for j in range(len(batch_tensors))
            ]
 
        # 2. Pad tensors to correct length
        new_batch_tensors = []
        for batch in batch_tensors:
            new_tensors = []
            # [field1, field2]
            for field_tensor in batch: # field1
                for i in range(len(self.dims)):
                    dim = self.dims[i]
                    target_length = target_lengths[i]
                    field_tensor = field_tensor.transpose(dim, -1)
                    field_tensor = torch.nn.functional.pad(
                        field_tensor, (
                            0, target_length - field_tensor.shape[-1]), value=self.values[i])    
                    field_tensor = field_tensor.transpose(-1, dim)
                new_tensors.append(field_tensor)
                # [pad_field1, pad_field2]
            new_batch_tensors.append(new_tensors) # [ [pad_field1, pad_field2] , [pad_field1, pad_field2] ]
        
        return new_batch_tensors, unpadded_lengths